/*
 Copyright 2015 Nicolas Melot, Johan Janzén

 This file is part of Drake-SCC.

 Drake-SCC is free software: you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation, either version 3 of the License, or
 (at your option) any later version.

 Drake-SCC is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with Drake-SCC. If not, see <http://www.gnu.org/licenses/>.

*/


#include <stdio.h>
#include <stdarg.h>
#include <semaphore.h>
#include <time.h>
#include <pthread.h>
#include <math.h>

#include <drake/platform.h>
#include <drake/stream.h>
#include <pelib/pelib_scc.h>
#include <pelib/scc_printf.h>
#include <pelib/malloc.h>

#ifdef COPPERRIDGE
#include "config.h"
#endif
#include <RCCE.h>
#include <RCCE_lib.h>
#include <RCCE_lib_pwr.h>

#include <pelib/sync.h>
#include <pelib/monitor.h>

#if 0
#define debug(var) printf("[%s:%s:%d:CORE %zu] %s = \"%s\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), #var, var); fflush(NULL)
#define debug_addr(var) printf("[%s:%s:%d:CORE %zu] %s = \"%p\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), #var, var); fflush(NULL)
#define debug_int(var) printf("[%s:%s:%d: CORE %zu] %s = \"%d\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), #var, var); fflush(NULL)
#define debug_size_t(var) printf("[%s:%s:%d: CORE %zu] %s = \"%zu\"\n", __FILE__, __FUNCTION__, __LINE__, drake_platform_core_id(), #var, var); fflush(NULL)
#else
#define debug(var)
#define debug_addr(var)
#define debug_int(var)
#define debug_size_t(var)
#endif

#define DRAKE_SCC_LINE_SIZE 32

typedef struct args
{
	size_t argc;
	char **argv;
} scc_args_t;

static void* alloc_ptr;

#ifndef SCC
#define VOLTAGE_LOW 3.296
#define VOLTAGE_HIGH 3.9
static int *frequency;
static float *voltage;
#endif

// Required for RCCE power and frequency scaling
extern t_vintp frequency_change_virtual_address[RCCE_MAXNP/2];
extern triple RC_V_MHz_cap[];
int RC_voltage_level(int Fdiv); //RCCE internal-ish function, side effect free.

#define CORES_IN_DOMAIN 8
#define DOMAINS 6
static const int power_domain_members_ue[DOMAINS][CORES_IN_DOMAIN] =
	{
		{ 0, 1, 2, 3,12,13,14,15},
		{ 4, 5, 6, 7,16,17,18,19},
		{ 8, 9,10,11,20,21,22,23},
		{24,25,26,27,36,37,38,39},
		{28,29,30,31,40,41,42,43},
		{32,33,34,35,44,45,46,47}
	};
      
/*
void power_handling_init(void)
{
  int my_domain = RCCE_power_domain();
  int i;
  int max_frequency = 100; // never lower than this on SCC
  for(i = 0; i < CORES_IN_DOMAIN; i++)
    {
      int ue = power_domain_members_ue[my_domain][i];
      if(ue < NO_CORES)
	{
	  int freq = max_frequency_all[ue];
	  //if(RCCE_ue() == 10) printf("(%i) checking %i,%i (%i). frequency is %i\n",RCCE_ue(),my_domain,i,power_domain_members_ue[my_domain][i],freq);
	  max_frequency = max_frequency < freq ? freq : max_frequency;
	}
    }
  //printf("(%i) My max frequency is  %i. my domain is %i\n",RCCE_ue(),max_frequency,my_domain);
  set_frequency(100); // min frequency, always safe
  set_DVFS(max_frequency);
  set_frequency(max_frequency);
}
*/

// Required for time measurement

struct drake_time {
	double time;
};
typedef struct drake_time *drake_time_t;

#if defined(__i386__)
// from http://stackoverflow.com/questions/9887839/clock-cycle-count-wth-gcc
__inline__ unsigned long long rdtsc(void)
{
    unsigned long long int x;
    __asm__ volatile (".byte 0x0f, 0x31" : "=A" (x));
    return x;
}
#elif defined(__x86_64__)
__inline__ unsigned long long rdtsc(void)
{
    unsigned hi, lo;
    __asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
    return ( (unsigned long long)lo)|( ((unsigned long long)hi)<<32 );
}
#endif

struct drake_platform
{
	drake_stream_t stream;
};
typedef struct drake_platform *drake_platform_t;

static volatile void **mpb;
static pelib_malloc_queue_t *shared;

volatile void*
remote_addr(volatile void* addr, size_t core)
{
	return pelib_scc_global_ptr((void*)addr, core);
}

drake_platform_t
drake_platform_init(void* obj)
{
	// Makes sure something has been transmitted
	if(obj == NULL)
	{
		fprintf(stderr, "Drake for SCC requires the commandline parameters to transmit to RCCE\n");
		//abort();
	}

	scc_args_t *args = (scc_args_t*) obj;
#ifdef SCC
	int freq = (int)(atof(args->argv[2]) * 1000 * 1000);
#endif
	size_t i;
	int int_argc = (int)args->argc;
	pelib_scc_init(&int_argc, &args->argv);
	args->argc = (size_t)int_argc;

	// Initialize memory allication
	drake_platform_t platform = malloc(sizeof(struct drake_platform));
	
	shared = (pelib_malloc_queue_t*)malloc(sizeof(pelib_malloc_queue_t) * RCCE_num_ues());
	mpb = (volatile void**)malloc(sizeof(void*) * RCCE_num_ues());
	mpb[drake_platform_core_id()] = RCCE_malloc(mpb_size - 32);
	for(i = 0; i < RCCE_num_ues(); i++)
	{
		if(i != drake_platform_core_id())
		{
			mpb[i] = remote_addr(mpb[drake_platform_core_id()], i);
		}
		pelib_mem_malloc_init(&shared[i], (void*)mpb[i], mpb_size - 32);
	}

	// Initiate redirection
	//pelib_scc_init_redirect();
	//pelib_scc_set_redirect();

#ifndef SCC
	frequency = malloc(sizeof(int) * RCCE_num_ues());
	voltage = malloc(sizeof(float) * RCCE_num_ues());

	for(i = 0; i < RCCE_num_ues(); i++)
	{
		frequency[i] = 100000;
		voltage[i] = VOLTAGE_LOW;
	}
#else
	drake_platform_set_voltage_frequency(freq);
#endif	

	return platform;
}

int
drake_platform_destroy(drake_platform_t stream)
{
	// Stop and finalize redirection
	//pelib_scc_finalize_redirect();
	//pelib_scc_stop_redirect();

	free(shared);
	RCCE_free(mpb[drake_platform_core_id()]);

	// Shutdown RCCE
	pelib_scc_finalize();

	// Free the stream container
	free(stream);

#ifndef SCC
	free(frequency);
	free(voltage);
#endif

	return 0;
}

int
drake_platform_stream_create_explicit(drake_platform_t stream, void (*schedule_init)(), void (*schedule_destroy)(), void* (*task_function)(size_t id, task_status_t status))
{
	stream->stream = drake_stream_create_explicit(schedule_init, schedule_destroy, task_function, stream);
	return 1;
}

int
drake_platform_stream_init(drake_platform_t stream, void* arg)
{
	return drake_stream_init(&stream->stream, arg);
}

int
drake_platform_stream_run(drake_platform_t stream)
{
	return drake_stream_run(&stream->stream);
}

int
drake_platform_stream_destroy(drake_platform_t stream)
{
	return drake_stream_destroy(&stream->stream);
}

volatile
void*
drake_platform_shared_malloc(size_t size, size_t core)
{
	return pelib_mem_malloc(&shared[core], size, DRAKE_SCC_LINE_SIZE);
}

volatile
void*
drake_platform_shared_malloc_mailbox(size_t size, size_t core)
{
	pelib_malloc_block_t *t = shared[core].tail;
	while(t->next != shared[core].tail)
	{
		t = t->next;
	}
	return pelib_mem_malloc(&shared[core], size, 0);
}

void
drake_platform_shared_free(volatile void* addr, size_t core)
{
	pelib_mem_free(&shared[core], (void*)addr);
}

void*
drake_platform_store_malloc(size_t size)
{
	return (void*)malloc(size);
}

void
drake_platform_store_free(void *addr)
{
	free(addr);
}

void*
drake_platform_private_malloc(size_t size)
{
	return malloc(size);
}

void
drake_platform_private_free(void *addr)
{
	free(addr);
}

int drake_platform_pull(volatile void* addr)
{
	(void*)(addr);
	pelib_scc_cache_invalidate();

	// This cannot go wrong
	return 1;
}

int drake_platform_commit(volatile void* addr)
{
	(void*)(addr);
	// Wherever you wrote, just invalidate the line
	pelib_scc_force_wcb();

	// This cannot go wrong
	return 1;
}

size_t
drake_platform_core_id()
{
	return RCCE_ue();
}

size_t
drake_platform_core_size()
{
	return RCCE_num_ues();
}

size_t
drake_platform_core_max()
{
	return 48;
}

void
drake_platform_barrier(void* channel)
{
	RCCE_barrier(&RCCE_COMM_WORLD);
}

void
drake_exclusive_begin()
{
	PELIB_SCC_CRITICAL_BEGIN
}

void
drake_exclusive_end()
{
	PELIB_SCC_CRITICAL_END
}

size_t
drake_platform_store_size()
{
	return STORE_SIZE;
}

size_t
drake_platform_shared_size()
{
	return mpb_size - 32;
}

size_t
drake_platform_shared_align()
{
	return DRAKE_SCC_LINE_SIZE;
}

static
int
frequency_to_rcce_div(int frequency_mhz)
{
	/*
	static const int frequency_to_rcce_div[][2] = {
	{800,2},
	{533,3},
	{400,4},
	{320,5},
	{266,6},
	{228,7},
	{200,8},
	{178,9},
	{160,10},
	{145,11},
	{133,12},
	{123,13},
	{114,14},
	{106,15},
	{100,16}};
	*/
	int frequency_div = RC_GLOBAL_CLOCK_MHZ / frequency_mhz;
	if(frequency_div < 17 && frequency_div > 1)// && RC_GLOBAL_CLOCK_MHZ % frequency_mhz == 0 ) // no
	{
		return frequency_div;
	}

	printf("Invalid frequency: %i\n",frequency_mhz);
	exit(1);
	return -1;
}

int
drake_platform_set_frequency(int freq /* in KHz */)
{
	if(freq < 1000)
	{
		fprintf(stderr, "Error: Requested frequency too low: %i Khz. Drake support frequency levels of at least 1KHz.\n", freq / 1000);
		return 0;
	}
#ifdef SCC
	int new_fdiv = frequency_to_rcce_div(freq / 1000);

	int Vlevel = RC_voltage_level(new_fdiv);
	if (RC_current_voltage_level < Vlevel || Vlevel < 0)
	{
		fprintf(stderr, "Error: current voltage level (%i) does not support requested frequency (%i Mhz). Required level is (%i)\n", RC_current_voltage_level, freq / 1000, Vlevel);
		return 0;
	}

	// At this point we know that the current voltage level is high enough 
	// (though not neccesarily optimal) for the requested frequency, so it is safe to change the frequency

	int my_rcce_tileid = RCCE_ue() / 2; //rounding down;
	*(frequency_change_virtual_address[my_rcce_tileid]) = FID_word(new_fdiv, my_rcce_tileid);
	//printf("IM ABOUT TO SET FREQUENCY. I want frequency divider %i, current voltage level is %i, needed level is %i\n",new_fdiv,RC_current_voltage_level,Vlevel);
	RC_current_frequency_divider = new_fdiv; // Be nice to RCCE.
#else
	frequency[RCCE_ue()] = freq;
#endif

	return 1;
}

int
drake_platform_set_frequency_autoscale(int frequency /* in KHz */)
{
	fprintf(stderr, "[%zu][ERROR] %s: not implemented.\n", drake_platform_core_id(), __FUNCTION__);
	return 0;
}

/* RCCE use the following values. see RC_V_MHz_cap[] in RCCE_power_management.c
* Frequency  Voltage "Voltagelevel"
* 800        1.1     4
* 533        0.8     1
* 400        0.8     0
*<400                0
*
* There are 7 voltage levels, but only 2 actual voltages are being used.
*/

int
drake_platform_set_voltage(float voltage /* in volts */)
{
	fprintf(stderr, "[%zu][ERROR] %s: not implemented.\n", drake_platform_core_id(), __FUNCTION__);
	return 0;
}

int
drake_platform_set_voltage_frequency(drake_platform_t pt, size_t freq /* in KHz */)
{
#ifdef SCC
	if(RC_current_voltage_level > 0)
	{
		int requested_Fdiv = frequency_to_rcce_div(freq / 1000);
		int new_Vlevel;
		int new_Fdiv;
		RCCE_REQUEST req;
		int retval = RCCE_iset_power(requested_Fdiv, &req, &new_Fdiv, &new_Vlevel);

		if(retval != RCCE_SUCCESS)
		{
			printf("Error: Something went wrong with voltage scaling\n");
			//exit(retval);
			return 0;
		}
		RCCE_wait_power(&req);
		// TODO: some cores may be already done and won't change frequency anymore
		//RCCE_barrier(&RCCE_COMM_WORLD);

		// RCCE only maintains these variables for the domain
		// masters, but we need them properly set for all processors.
		// By doing this, we assume the domain masters did their job.
		RC_current_frequency_divider = new_Fdiv;
		RC_current_voltage_level = RC_voltage_level(requested_Fdiv);
#else
	if(frequency[RCCE_ue()] > 0)
	{
		frequency[RCCE_ue()] = freq;
		if(frequency[RCCE_ue()] >= 800000)
		{
			voltage[RCCE_ue()] = VOLTAGE_HIGH;
		}
		else
		{
			voltage[RCCE_ue()] = VOLTAGE_LOW;
		}
#endif
	}

	return 1;
}

size_t
drake_platform_get_frequency(drake_platform_t pt) /* in KHz */
{
#ifdef SCC
	if(RC_current_frequency_divider > 0)
	{
		return ((int)(RC_GLOBAL_CLOCK_MHZ / RC_current_frequency_divider)) * 1000;
#else
	if(frequency[RCCE_ue()] > 0)
	{
		return frequency[RCCE_ue()];
#endif
	}
	else
	{
		return 0;
	}
}

float
drake_platform_get_voltage() /* in volts */
{
#ifdef SCC
	return RC_V_MHz_cap[RC_current_voltage_level].volt;
#else
	return voltage[RCCE_ue()];
#endif
}

int drake_platform_time_get(drake_time_t container)
{
	struct timespec t;
	clock_gettime(CLOCK_MONOTONIC, &t);
	container->time = t.tv_sec * 1e3 + t.tv_nsec / 1e6;
	//rdtsc() / drake_platform_get_frequency();
	return 1;
}

int drake_platform_time_subtract(drake_time_t res, drake_time_t t1, drake_time_t t2)
{
	res->time = t1->time - t2->time;
	return 1;
}

int drake_platform_time_add(drake_time_t res, drake_time_t t1, drake_time_t t2)
{
	res->time = t1->time + t2->time;
	return 1;
}

int drake_platform_time_greater(drake_time_t t1, drake_time_t t2)
{
	//printf("[%s:%s:%d] t1 = %F > t2 = %F.\n", __FILE__, __FUNCTION__, __LINE__, t1->time, t2->time);
	return t1->time > t2->time;
}

int drake_platform_time_equals(drake_time_t t1, drake_time_t t2)
{
	return t1->time == t2->time;
}

int drake_platform_time_init(drake_time_t t, double ms)
{
	t->time = ms;
	return 1;
}

int drake_platform_sleep(drake_time_t period)
{
	// Todo: improve with a proper sleep() system call
	struct drake_time now, until;
	drake_time_t now_p = &now, until_p = &until;

	drake_platform_time_get(now_p);
	until_p->time = now_p->time + period->time;
	//printf("[%s:%s:%d] sleeping for %Fms from t1 = %F to t2 = %F.\n", __FILE__, __FUNCTION__, __LINE__, period->time, now_p->time, until_p->time);
	while(!drake_platform_time_greater(now_p, until_p))
	{
		drake_platform_time_get(now_p);
	}

	return 1;
}

drake_time_t
drake_platform_time_alloc()
{
	return malloc(sizeof(drake_time_t));
}

FILE*
drake_platform_time_printf(FILE *stream, drake_time_t time)
{
	fprintf(stream, "%f", time->time);
	return stream;
}

void
drake_platform_time_destroy(drake_time_t time)
{
	free(time);
}

struct drake_power {
	double *power_chip, *power_core, *power_mc, *time;
	size_t max_samples;
	size_t collected;
	sem_t run;
	int running;
	int measurement;
	pthread_t thread;
};
typedef struct drake_power *drake_power_t;

// Measurement enabled for SCC_3V3 (fixed supply voltage for chip. Not really expected to change, but might fluctuate slightly), 
// MC_1V5 (memory controller voltage ), 
// SCC3V3C (chip current)
// MC_1V5C (memory controller current)
#define DVFS_CONFIG_SETTING 0x00000F80 
#define DVFS_CONFIG_RESET 0x00000000 // Measurement reset to default setting
#define REASONABLE_C 100.0
#define REASONABLE_MC 50.0 // I dont know

void*
measure_power(void* arg)
{
	drake_power_t args = (drake_power_t)arg;
	const struct timespec period = {0, 100 * 1e6}; // 100 ms

	// Wait for kickstarter
	sem_wait(&args->run);

#ifdef COPPERRIDGE
#warning Do not forget to set "i2c off" from the bmc console (telnet 10.3.16.164 5010) before running this program
	InitAPI(0);  
	writeFpgaGrb(DVFS_CONFIG, DVFS_CONFIG_SETTING);
#endif

	const struct timespec short_sleeptime = {0, 5 * 1e6}; // 5 ms
	double time_ms;

	struct timespec time;
	clock_gettime(CLOCK_MONOTONIC, &time);
	double start_ms = time.tv_sec * 1e3 + time.tv_nsec * 1e-6;

	args->running = 1;
	while(sem_trywait(&args->run) != 0)
	{
		if(args->collected < args->max_samples)
		{
			if((args->measurement & (1 << DRAKE_POWER_CHIP)) != (int)0)
			{
#ifdef COPPERRIDGE
				double voltage, current;
				voltage = readStatus(DVFS_STATUS_U3V3SCC);
				current = readStatus(DVFS_STATUS_I3V3SCC);
				args->power_chip[args->collected] = voltage * current;
#else
				float mean_v = 0;
				int mean_f = 0;
				size_t i;
				for(i = 0; i < RCCE_num_ues(); i++)
				{
					mean_v += voltage[i];
					mean_f += frequency[i];
				}
				mean_v = mean_v / RCCE_num_ues();
				mean_f = mean_f / RCCE_num_ues();

				args->power_chip[args->collected] = pow(mean_v, 2) * mean_f;
#endif
			}
			else
			{
				args->power_chip[args->collected] = 0;
			}

			if((args->measurement & (1 << DRAKE_POWER_MEMORY_CONTROLLER)) != (int)0)
			{
#ifdef COPPERRIDGE
				double voltage, current;
				voltage = readStatus(DVFS_STATUS_U1V5);
				current = readStatus(DVFS_STATUS_I1V5);
				args->power_mc[args->collected] = voltage * current;
#else
				args->power_mc[args->collected] = 0;
#endif
			}
			else
			{
				args->power_mc[args->collected] = 0;
			}
     
			// No scc support for core-per-core power measurement 
			args->power_core[args->collected] = 0;

#ifdef COPPERRIDGE
			// There are some unreasonably high spikes every 500-700 ms or so. We ignore these and wait for the value to settle.
			if(args->power_chip[args->collected] > REASONABLE_C || args->power_mc[args->collected] > REASONABLE_MC)
			{
				nanosleep(&short_sleeptime, NULL);
				continue;
			}
#endif

			clock_gettime(CLOCK_MONOTONIC,&time);
			time_ms = time.tv_sec * 1e3 +time.tv_nsec * 1e-6 - start_ms;
			args->time[args->collected] = time_ms;
		}

		// Sleep until next measurement
		nanosleep(&period, NULL);

		// Count the number of iterations
		args->collected++;
	}

	args->running = 0;
#ifdef COPPERRIDGE
	writeFpgaGrb(DVFS_CONFIG, DVFS_CONFIG_RESET); // Go back to no measurement
#endif

	return NULL;
}

drake_power_t
drake_platform_power_init(drake_platform_t pt, size_t samples, int measurement)
{
	drake_power_t tracker = malloc(sizeof(struct drake_power));
	tracker->power_chip = malloc(sizeof(double) * samples);
	tracker->power_mc = malloc(sizeof(double) * samples);
	tracker->power_core = malloc(sizeof(double) * samples);
	tracker->time = malloc(sizeof(double) * samples);
	double *stuff = malloc(sizeof(double) * samples);
	tracker->max_samples = samples;
	tracker->collected = 0;
	tracker->running = 0;
	tracker->measurement = measurement;

	sem_init(&tracker->run, 0, 0);
	int retval = 0;
	//retval = pthread_create(&(tracker->thread), NULL, measure_power, tracker);
	if(retval != 0)
	{
		free(tracker->power_core);
		free(tracker->power_mc);
		free(tracker->time);
		free(tracker);
		tracker = NULL;
	}

	return tracker;
}

void
drake_platform_power_begin(drake_power_t tracker)
{
	sem_post(&tracker->run);
}

size_t
drake_platform_power_end(drake_power_t tracker)
{
	sem_post(&tracker->run);
  	pthread_join(tracker->thread, NULL);
	return tracker->collected;
}

FILE*
drake_platform_power_printf_line_cumulate(FILE* stream, drake_power_t tracker, size_t i, int metrics, char *separator)
{
	double line = 0;
	if((metrics & (1 << DRAKE_POWER_CHIP)) != 0)
	{
		line += tracker->power_chip[i];
	}

	if((metrics & (1 << DRAKE_POWER_MEMORY_CONTROLLER)) != 0)
	{
		line += tracker->power_mc[i];
	}

	if((metrics & (1 << DRAKE_POWER_CORE)) != 0)
	{
		line += tracker->power_core[i];
	}

	fprintf(stream, "%f%s%f", tracker->time[i], separator, line);
	return stream;
}

FILE*
drake_platform_power_printf_cumulate(FILE* stream, drake_power_t tracker, int metrics, char *separator)
{
	size_t i;
	for(i = 0; i < (tracker->collected > tracker->max_samples ? tracker->max_samples : tracker->collected); i++)
	{
		drake_platform_power_printf_line_cumulate(stream, tracker, i, metrics, separator);
	}

	return stream;
}

FILE*
drake_platform_power_printf_line(FILE* stream, drake_power_t tracker, size_t i, char* separator)
{
	if(separator == NULL)
	{
		separator = " ";
	}

	fprintf(stream, "%f%s", tracker->time[i], separator);
	if(tracker->measurement & (1 << DRAKE_POWER_CHIP) != 0)
	{
		fprintf(stream, "%f", tracker->power_chip[i]);
		
		// If there is anything else than power chip, then add a separator
		if(((tracker->measurement >> DRAKE_POWER_CHIP) << DRAKE_POWER_CHIP) & ~(int)(1 << DRAKE_POWER_CHIP) != 0)
		{
			fprintf(stream, "%s", separator);
		}
	}

	if(tracker->measurement & (1 << DRAKE_POWER_MEMORY_CONTROLLER) != 0)
	{
		fprintf(stream, "%f", tracker->power_mc[i]);
		if(((tracker->measurement >> DRAKE_POWER_MEMORY_CONTROLLER) << DRAKE_POWER_MEMORY_CONTROLLER) & ~(int)(1 << DRAKE_POWER_MEMORY_CONTROLLER) != 0)
		{
			fprintf(stream, "%s", separator);
		}
	}

	if(tracker->measurement & (1 << DRAKE_POWER_CORE) != 0)
	{
		fprintf(stream, "%f", tracker->power_core[i]);
		if(((tracker->measurement >> DRAKE_POWER_CORE) << DRAKE_POWER_CORE) & ~(int)(1 << DRAKE_POWER_CORE) != 0)
		{
			fprintf(stream, "%s", separator);
		}
	}

	return stream;
}

FILE*
drake_platform_power_printf(FILE* stream, drake_power_t tracker, char *separator)
{
	if(separator == NULL)
	{
		separator = " ";
	}

	size_t i;
	for(i = 0; i < (tracker->collected > tracker->max_samples ? tracker->max_samples : tracker->collected); i++)
	{
		drake_platform_power_printf_line(stream, tracker, i, separator);
		fprintf(stream, "\n");
	}

	return stream;
}

void
drake_platform_power_destroy(drake_power_t tracker)
{
	if(tracker->running != 0)
	{
		drake_platform_power_end(tracker);
	}

	free(tracker->power_chip);
	free(tracker->power_mc);
	free(tracker->power_core);
	free(tracker->time);
	free(tracker);
}

void
drake_platform_sleep_enable(drake_platform_t pt, size_t core)
{
	// Do nothing
}

void
drake_platform_core_disable(drake_platform_t pt, size_t core)
{
	// Do nothing
}

